/*
 * This file is part of queso.
 *
 * Copyright © 2020 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace Queso {

    /**
     * Class for Queso windows.
     */
    [GtkTemplate (ui = "/mx/unam/Queso/application-window.ui")]
    public class ApplicationWindow : Gtk.ApplicationWindow {

        /* Record icon name. */
        private const string RECORD_NAME = "media-record-symbolic";
        /* Stop icon name. */
        private const string STOP_NAME = "media-playback-stop-symbolic";
        /* Icon size. */
        private const Gtk.IconSize ICON_SIZE = Gtk.IconSize.SMALL_TOOLBAR;


        /* The header bar. */
        [GtkChild]
        private Gtk.HeaderBar header;

        [GtkChild]
        private Gtk.ToggleButton video_source_button;
        [GtkChild]
        private Gtk.MenuButton microphone_button;
        [GtkChild]
        private Gtk.Popover microphone_popover;
        [GtkChild]
        private Gtk.ComboBoxText microphone_combo_box;


        /* The elements entry. */
        [GtkChild]
        private Gtk.Image record_stop_image;

        /* The application. */
        private Application app;
        private Devices devices;
        private VideoSourceSettingsDialog vs_settings_dialog;

        /**
         * Constructs a new application window.
         * @param application the application.
         */
        public ApplicationWindow(Gtk.Application application) {
            GLib.Object(application: application);
            app = application as Application;
            app.event_ocurred.connect(event_ocurred);

            // camera_button.bind_property(
            //     "active", camera_popover,
            //     "visible", GLib.BindingFlags.BIDIRECTIONAL);
            // microphone_button.bind_property(
            //     "active", microphone_popover,
            //     "visible", GLib.BindingFlags.BIDIRECTIONAL);
        }

        /**
         * Adds the camera widget to the window.
         * @param widget the camera widget.
         * @param width the widget width.
         * @param height the height width.
         */
        public void add_camera_widget(Gtk.Widget widget,
                                      int width, int height) {
            this.add(widget);
            widget.width_request = width;
            widget.height_request = height;
            widget.visible = true;
        }

        public void add_devices(Devices devices) {
            this.devices = devices;
            
            // foreach (var device in devices.video_devices) {
            //     camera_combo_box.append_text(device.name);
            // }
            // foreach (var device in devices.audio_devices) {
            //     microphone_combo_box.append_text(device.name);
            // }
            // camera_combo_box.active_id = devices.video_device.name;
            // microphone_combo_box.active_id = devices.audio_device.name;
        }

        /* The on close activated event handler. */
        [GtkCallback]
        private bool on_close_activated() {
            app.activate_action("quit", null);
            return true;
        }

        /* The on record/stop clicked event handler. */
        [GtkCallback]
        private void on_record_stop_clicked() {
            app.activate_action("toggle-record", null);
        }

        /* The on camera combo changed event handler. */
        [GtkCallback]
        private void on_video_source_button_toggled() {
            //app.activate_action("video-source-settings", null);
            toggle_video_source_settings();
        }

        private bool dispose_video_settings_dialog() {
            video_source_button.active = false;
            return GLib.Source.REMOVE;
        }

        /* The on microphone combo changed event handler. */
        [GtkCallback]
        private void on_microphone_combo_changed() {
            app.change_audio_device(microphone_combo_box.active_id);
        }

        /* Event ocurred handler. */
        private void event_ocurred(QuesoEvent event) {
            switch (event) {
            case RECORDING_TOGGLED:
                recording_toggled();
                break;
            case VIDEO_SOURCE_SETTINGS:
                video_source_button.active = !video_source_button.active;
                break;
            }
        }

        private void toggle_video_source_settings() {
            if (video_source_button.active) {
                vs_settings_dialog = (
                    new VideoSourceSettingsDialog(this, devices));
                vs_settings_dialog.show_all();
                vs_settings_dialog.delete_event.connect(
                    dispose_video_settings_dialog);
            } else {
                if (vs_settings_dialog != null)
                    vs_settings_dialog.dispose();
                vs_settings_dialog = null;
            }
        }

        /* Recording toggled event handler. */
        private void recording_toggled() {
            if (!app.is_recording) {
                record_stop_image.set_from_icon_name(RECORD_NAME, ICON_SIZE);
                header.subtitle = "Webcam";
            } else {
                record_stop_image.set_from_icon_name(STOP_NAME, ICON_SIZE);
                header.subtitle = "Recording";
            }
        }

        public void audio_device_changed(AudioDevice new_audio_device) {
            if (microphone_combo_box.active_id == new_audio_device.name)
                return;
            microphone_combo_box.active_id = new_audio_device.name;
        }

        // public void video_device_changed(VideoDevice new_video_device) {
        //     if (camera_combo_box.active_id == new_video_device.name)
        //         return;
        //     camera_combo_box.active_id = new_video_device.name;
        // }
    }
}
