/*
 * This file is part of queso.
 *
 * Copyright © 2020 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace Queso {

    /**
     * Class for audio devices.
     */
    public class AudioDevice : Device, Gee.Comparable<AudioDevice> {

        /**
         * The device setups.
         */
        public Gee.TreeSet<AudioSetup> setups { public get; private set; }

        /**
         * Constructs an audio device.
         * @param device the GStreamer device.
         */
        public AudioDevice(Gst.Device device) {
            base(device);
        }

        /**
         * Constructs a testing video device.
         */
        public AudioDevice.test() {
            base.test();
            name = "Test Audio Source";
            setups.add(new AudioSetup(AudioSetup.RAW,
                                      new Gee.TreeSet<Format>(),
                                      null, 1, 2147483647, 1, 2));
        }

        /**
         * Initializes the audio device setups.
         */
        protected override void init_setups() {
            setups = new Gee.TreeSet<AudioSetup>();
        }

        /**
         * Parse an structure in the device.
         * @param structure the structure to parse.
         */
        protected override void parse_structure(Gst.Structure structure) {
            string n = "";
            string l = null;
            int r1 = -1;
            int r2 = -1;
            int c1 = -1;
            int c2 = -1;
            n = structure.get_name();
            if (n == null)
                return;
            var fmts = new Gee.TreeSet<Format>();
            l = structure.get_string("layout");
            GLib.Value? f = structure.get_value("format");
            GLib.Value? r = structure.get_value("rate");
            GLib.Value? c = structure.get_value("channels");
            if (f != null && f.type() == typeof(Gst.ValueList)) {
                for (int i = 0; i < Gst.ValueList.get_size(f); i ++) {
                    var vf = Gst.ValueList.get_value(f, i);
                    if (vf.type() != typeof(string))
                        continue;
                    fmts.add(new Format(vf.dup_string()));
                }
            } else if (f != null && f.type() == typeof(string)) {
                fmts.add(new Format(f.dup_string()));
            }
            if (r.type() == typeof(Gst.IntRange))
                parse_range(r, out r1, out r2);
            if (c.type() == typeof(Gst.IntRange))
                parse_range(c, out c1, out c2);
            setups.add(new AudioSetup(n, fmts, l, r1, r2, c1, c2));
        }

        /* Parses a range. */
        private void parse_range(GLib.Value range, out int min, out int max) {
            min = Gst.Value.get_int_range_min(range);
            max = Gst.Value.get_int_range_max(range);
        }

        /**
         * Compares the two audio devices.
         * @param dev the audio device to compare with.
         * @return -1 if this device it's better than the one received; 0 if
         *         they are equal; 1 if it's worse.
         */
        public int compare_to(AudioDevice dev) {
            if (!is_test && dev.is_test)
                return -1;
            if (is_test && !dev.is_test)
                return 1;
            var ts = this.setups.first();
            var ds = dev.setups.first();
            return ts.compare_to(ds);
        }

        public new Gst.Element create_element(string name) {
            if (is_test)
                return Gst.ElementFactory.make("testaudiosrc", name);
            return base.create_element(name);
        }
    }
}
