/*
 * This file is part of queso.
 *
 * Copyright © 2020 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace Queso {

    /**
     * Class for audio setups.
     */
    public class AudioSetup : GLib.Object, Gee.Comparable<AudioSetup> {

        /**
         * Audio raw format.
         */
        public const string RAW = "audio/x-raw";

        /**
         * The media type.
         */
        public string media_type { public get; private set; }

        /**
         * The formats.
         */
        public Gee.TreeSet<Format> formats { public get; private set; }

        /**
         * The layout.
         */
        public string layout { public get; private set; }

        /**
         * The minimum rate.
         */
        public int min_rate { public get; private set; }

        /**
         * The maximum rate.
         */
        public int max_rate { public get; private set; }

        /**
         * The minimum channel.
         */
        public int min_channel { public get; private set; }

        /**
         * The maximum channel.
         */
        public int max_channel { public get; private set; }

        /**
         * Constructs an audio setup.
         */
        public AudioSetup(string media_type, Gee.TreeSet<Format> formats,
                          string? layout = null,
                          int min_rate = -1, int max_rate = -1,
                          int min_channel = -1, int max_channel = -1) {
            this.media_type = media_type;
            this.formats = formats;
            this.layout = layout;
            this.min_rate = min_rate;
            this.max_rate = max_rate;
            this.min_channel = min_channel;
            this.max_channel = max_channel;
        }

        /**
         * Returns a string representation of the audio setup.
         * @return a string representation of the audio setup.
         */
        public string to_string() {
            string l = (layout == null) ? "" :
                "layout=(string)%s".printf(layout);
            string r = min_rate < 1 ||  max_rate < 1 ? "" :
                "rate=(int)[ %d, %d ]".printf(min_rate, max_rate);
            string c = min_channel < 1 || max_channel < 1 ? "" :
                "channels=(int)[ %d, %d ]".printf(min_channel, max_channel);
            string sfmts = "";
            if (formats.size == 1) {
                var fmt = formats.first();
                sfmts = "format=(string)%s".printf(fmt.format);
            } else {
                foreach (var fmt in formats) {
                    if (sfmts == "") {
                        sfmts = "format=(string){ %s".printf(fmt.format);
                    } else {
                        sfmts += ", %s".printf(fmt.format);
                    }
                }
                if (!formats.is_empty)
                    sfmts += " }";
            }
            string rs = media_type;
            if (sfmts != "")
                rs += ", %s".printf(sfmts);
            if (l != "")
                rs += ", %s".printf(l);
            if (r != "")
                rs += ", %s".printf(r);
            if (c != "")
                rs += ", %s".printf(c);
            return rs;
        }

        /**
         * Compares the two audio setups.
         * @param setup the audio setup to compare with.
         * @return -1 if this setup it's better than the one received; 0 if
         *         they are equal; 1 if it's worse.
         */
        public int compare_to(AudioSetup setup) {
            if (this.media_type == RAW && setup.media_type != RAW)
                return -1;
            if (this.media_type != RAW && setup.media_type == RAW)
                return 1;
            if (!this.formats.is_empty && setup.formats.is_empty)
                return -1;
            if (this.formats.is_empty && !setup.formats.is_empty)
                return 1;
            if (!this.formats.is_empty && !setup.formats.is_empty) {
                var tf = this.formats.first();
                var af = setup.formats.first();
                return tf.compare_to(af);
            }
            if (this.max_rate != -1 && setup.max_rate == -1)
                return -1;
            if (this.max_rate == -1 && setup.max_rate != -1)
                return 1;
            if (this.max_rate != -1 && setup.max_rate != -1 &&
                this.max_rate != setup.max_rate)
                return this.max_rate - setup.max_rate;
            if (this.max_channel != -1 && setup.max_channel == -1)
                return -1;
            if (this.max_channel == -1 && setup.max_channel != -1)
                return 1;
            if (this.max_channel != -1 && setup.max_channel != -1 &&
                this.max_channel != setup.max_channel)
                return this.max_channel - setup.max_channel;
            return GLib.strcmp(this.media_type, setup.media_type);
        }
    }
}
