/*
 * This file is part of queso.
 *
 * Copyright © 2020 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace Queso {

    /**
     * Class to handle the pipeline.
     *
     * Raw idle pipeline:
     *
     *{{{
     *                          ┌─────────────────────┐
     *                          |      video-src      |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |    video-filter     |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |   video-queue-gtk   |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |  video-convert-gtk  |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |  video-text-record  |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |   video-text-info   |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |   video-gtk-sink    |
     *                          └─────────────────────┘
     * }}}
     *
     * Raw recording pipeline:
     *
     * {{{
     *             ┌─────────────────────┐             ┌─────────────────────┐
     *             |      video-src      |             |      audio-src      |
     *             └──────────┬──────────┘             └──────────┬──────────┘
     *                        │                                   │
     *             ┌──────────┴──────────┐             ┌──────────┴──────────┐
     *             |    video-filter     |             │    audio-filter     │
     *             └──────────┬──────────┘             └──────────┬──────────┘
     *                        │                                   │
     *             ┌──────────┴──────────┐             ┌──────────┴──────────┐
     *             | video-queue-pretee  |             │  audio-queue-rate   │
     *             └──────────┬──────────┘             └──────────┬──────────┘
     *                        │                                   |
     *             ┌──────────┴──────────┐             ┌──────────┴──────────┐
     *             |      video-tee      │             |     audio-rate      │
     *             └──────────┬──────────┘             └──────────┬──────────┘
     *            ┌───────────┴───────────┐                       │
     * ┌──────────┴──────────┐ ┌──────────┴──────────┐ ┌──────────┴──────────┐
     * |   video-queue-gtk   │ │  video-queue-file   │ │    audio-convert    │
     * └──────────┬──────────┘ └──────────┬──────────┘ └──────────┬──────────┘
     *            │                       │                       │
     * ┌──────────┴──────────┐ ┌──────────┴──────────┐ ┌──────────┴──────────┐
     * |  video-convert-gtk  | | video-convert-file  | │   audio-encoder     │
     * └──────────┬──────────┘ └──────────┬──────────┘ └──────────┬──────────┘
     *            │                       │                       │
     * ┌──────────┴──────────┐ ┌──────────┴──────────┐ ┌──────────┴──────────┐
     * |  video-text-record  | │ video-h264-encoder  │ │ audio-queue-mp4mux  │
     * └──────────┬──────────┘ └──────────┬──────────┘ └──────────┬──────────┘
     *            │                       └───────────┬───────────┘
     * ┌──────────┴──────────┐             ┌──────────┴──────────┐
     * |   video-text-info   |             |     file-mp4mux     |
     * └──────────┬──────────┘             └──────────┬──────────┘
     *            │                                   │
     * ┌──────────┴──────────┐             ┌──────────┴──────────┐
     * |   video-gtk-sink    |             |     file-queue      │
     * └─────────────────────┘             └──────────┬──────────┘
     *                                                │
     *                                     ┌──────────┴──────────┐
     *                                     |      file-sink      │
     *                                     └─────────────────────┘
     * }}}
     *
     * H.264 idle pipeline:
     *
     *{{{
     *                          ┌─────────────────────┐
     *                          |      video-src      |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |    video-filter     |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |  video-queue-parse  |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |  video-h264-parse   |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |   video-queue-gtk   |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          | video-h264-decoder  |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          | video-queue-convert |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |  video-convert-gtk  |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |  video-text-record  |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |   video-text-info   |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |   video-gtk-sink    |
     *                          └─────────────────────┘
     * }}}
     *
     * H.264 recording pipeline:
     *
     * {{{
     *
     *                                                 ┌─────────────────────┐
     *                                                 |      audio-src      |
     *                                                 └──────────┬──────────┘
     *                                                            │
     *             ┌─────────────────────┐             ┌──────────┴──────────┐
     *             |      video-src      |             |     audio-filter    |
     *             └──────────┬──────────┘             └──────────┬──────────┘
     *                        │                                   │
     *             ┌──────────┴──────────┐             ┌──────────┴──────────┐
     *             |    video-filter     |             │  audio-queue-rate   │
     *             └──────────┬──────────┘             └──────────┬──────────┘
     *                        │                                   │
     *             ┌──────────┴──────────┐             ┌──────────┴──────────┐
     *             |  video-queue-parse  |             │     audio-rate      │
     *             └──────────┬──────────┘             └──────────┬──────────┘
     *                        │                                   |
     *             ┌──────────┴──────────┐             ┌──────────┴──────────┐
     *             |  video-h264-parse   |             |    audio-convert    │
     *             └──────────┬──────────┘             └──────────┬──────────┘
     *                        │                                   │
     *             ┌──────────┴──────────┐             ┌──────────┴──────────┐
     *             |      video-tee      |             │    audio-encoder    │
     *             └──────────┬──────────┘             └──────────┬──────────┘
     *            ┌───────────┴───────────┐                       │
     * ┌──────────┴──────────┐ ┌──────────┴──────────┐ ┌──────────┴──────────┐
     * |   video-queue-gtk   │ │  video-queue-file   │ │ audio-queue-mp4mux  │
     * └──────────┬──────────┘ └──────────┬──────────┘ └──────────┬──────────┘
     *            │                       └───────────┬───────────┘
     * ┌──────────┴──────────┐             ┌──────────┴──────────┐
     * | video-h264-decoder  |             |     file-mp4mux     |
     * └──────────┬──────────┘             └──────────┬──────────┘
     *            │                                   │
     * ┌──────────┴──────────┐             ┌──────────┴──────────┐
     * | video-queue-convert |             │      file-sink      │
     * └──────────┬──────────┘             └─────────────────────┘
     *            │
     * ┌──────────┴──────────┐
     * |  video-convert-gtk  |
     * └──────────┬──────────┘
     *            │
     * ┌──────────┴──────────┐
     * |  video-text-record  |
     * └──────────┬──────────┘
     *            │
     * ┌──────────┴──────────┐
     * |   video-text-info   |
     * └──────────┬──────────┘
     *            │
     * ┌──────────┴──────────┐
     * |   video-gtk-sink    |
     * └─────────────────────┘
     * }}}
     *
     */
    public class Camera : GLib.Object {

        public const int DEFAULT_AUDIO_RATE = 44100;

        private Gtk.Widget _widget;
        public unowned Gtk.Widget widget {
            get {
                video_gtk_sink.get("widget", out _widget);
                return _widget;
            }
        }

        public bool saving { public get; private set; }

        /* The video and audio devices. */
        private Devices devices;
        private VideoDevice videodev;
        private VideoDevice new_videodev;
        private AudioDevice audiodev;
        private VideoSetup video_setup;
        private VideoSetup new_video_setup;
        private AudioSetup audio_setup;

        /* Pipeline and elements. */
        protected Gst.Pipeline pipeline;
        private Gst.Bus bus;

        /* Raw pipeline. */
        private Gst.Element video_src;
        private Gst.Element video_filter;
        private Gst.Element video_queue_decoder;
        private Gst.Element video_queue_convert;
        private Gst.Element video_queue_gtk;
        private Gst.Element video_convert_gtk;
        private Gst.Element video_text_record;
        private Gst.Element video_text_info;
        private Gst.Element video_gtk_sink;

        /* H.264 elements. */
        private Gst.Element video_queue_parse;
        private Gst.Element video_h264_parse;
        private Gst.Element video_h264_decoder;

        /* Audio pipeline. */
        private Gst.Element audio_src;
        private Gst.Element audio_filter;
        private Gst.Element audio_queue_rate;
        private Gst.Element audio_rate;
        private Gst.Element audio_convert;
        private Gst.Element audio_encoder;
        private Gst.Element audio_queue_mp4mux;

        /* File sink branch. */
        private Gst.Element video_tee;
        private Gst.Element video_queue_file;
        private Gst.Element video_convert_file;
        private Gst.Element video_h264_encoder;
        private Gst.Element file_mp4mux;

        /* File sink*/
        //private Gst.Element file_queue;
        private Gst.Element file_sink;

        /* Temporary filename . */
        private string temporary_fn;

        /* Timestamps. */
        private double start_saving_ts;
        private double last_saving_ts;

        private bool show_info;
        private int debug_counter;
        private bool last_change_done;

        public Camera(Devices devices) {
            this.devices = devices;
            devices.device_changed.connect(device_changed);
            videodev = devices.video_device;
            audiodev = devices.audio_device;
            video_setup = videodev.setups.first();
            audio_setup = audiodev.setups.first();

            pipeline = new Gst.Pipeline("queso");

            switch (video_setup.media_type) {
            case VideoSetup.RAW:
                raw_idle_pipeline();
                break;
            case VideoSetup.H264:
                h264_idle_pipeline();
                break;
            }

            bus = pipeline.get_bus();
            bus.add_watch(GLib.Priority.DEFAULT, pipeline_bus_watch);
            last_change_done = true;
        }

        /**
         * Starts the camera.
         */
        public void start() {
            GLib.Idle.add(start_camera);
        }

        /**
         * Stops the camera.
         */
        public void stop() {
            GLib.Idle.add(stop_camera);
        }

        /**
         * Cleans the camera and stops the pipeline.
         */
        public void clean() {
            pipeline.send_event(new Gst.Event.eos());
            if (file_sink != null)
                file_sink.set_state(Gst.State.NULL);
            if (file_mp4mux != null)
                file_mp4mux.set_state(Gst.State.NULL);
            if (saving) {
                var destination_fn = get_destination();
                move_file(temporary_fn, destination_fn);
                destination_fn = temporary_fn = null;
            }
        }

        /**
         * Toggles saving the camera video.
         */
        public void toggle_save() {
            saving = !saving;
            if (saving) {
                start_saving();
            } else {
                stop_saving();
            }
        }

        /**
         * Toggles the device information in the feed.
         */
        public void toggle_info() {
            show_info = !show_info;
            if (show_info) {
                video_text_info.set("text", get_info());
            } else {
                video_text_info.set("text", "");
            }
        }

        /**
         * Stores a dot file with the current pipeline, in the directory defined
         * by ''GST_DEBUG_DUMP_DOT_DIR'', with the name
         * ''queso-pipeline-XXX.dot.
         */
        public void debug() {
            var output = "queso-pipeline-%03d".printf(debug_counter++);
            Gst.Debug.bin_to_dot_file(pipeline, Gst.DebugGraphDetails.ALL,
                                      output);
        }

        /* Pipeline bus watch. */
        private bool
        pipeline_bus_watch(Gst.Bus bus, Gst.Message message) {
            switch (message.type) {
            case Gst.MessageType.ERROR:
                debug();
                GLib.Error error;
                string debug;
                message.parse_error(out error, out debug);
                GLib.warning("ERROR: " + error.message + "\n" + debug);
                return GLib.Source.CONTINUE;
            case Gst.MessageType.WARNING:
                debug();
                GLib.Error error;
                string debug;
                message.parse_warning(out error, out debug);
                GLib.warning("WARNING: " + error.message + "\n" + debug);
                return GLib.Source.CONTINUE;
            case Gst.MessageType.INFO:
                debug();
                GLib.Error error;
                string debug;
                message.parse_info(out error, out debug);
                GLib.warning("INFO: " + error.message + "\n" + debug);
                return GLib.Source.CONTINUE;
            case Gst.MessageType.ELEMENT:
                unowned Gst.Structure structure = message.get_structure();
                if (!structure.has_name("GstBinForwarded"))
                    return GLib.Source.CONTINUE;
                Gst.Message forward_message = null;
                structure.get("message", typeof(Gst.Message),
                              out forward_message);
                if (forward_message.type != Gst.MessageType.EOS)
                    return GLib.Source.CONTINUE;

                if (first_eos) {
                    stderr.printf("FIRST EOS!\n");
                    file_sink.set_state(Gst.State.NULL);
                    file_mp4mux.set_state(Gst.State.NULL);
                    temporary_fn = random_filename();
                    stderr.printf("SAVING TO %s\n", temporary_fn);
                    file_sink.set("location", temporary_fn);
                    file_sink.set_state(Gst.State.PLAYING);
                    file_mp4mux.set_state(Gst.State.PLAYING);
                    GLib.Idle.add(real_start_saving);
                    return GLib.Source.CONTINUE;
                }

                stderr.printf("EOS!\n");

                switch (video_setup.media_type) {
                case VideoSetup.RAW:
                    raw_eos();
                    break;
                case VideoSetup.H264:
                    h264_eos();
                    break;
                }

                var destination_fn = get_destination();
                move_file(temporary_fn, destination_fn);
                temporary_fn = destination_fn = null;
                return GLib.Source.CONTINUE;
            default:
                return GLib.Source.CONTINUE;
            }
        }

        /* Closes the file for a raw saving. */
        private void raw_eos() {
                video_tee.set_state(Gst.State.NULL);
                video_queue_gtk.set_state(Gst.State.NULL);
                video_queue_file.set_state(Gst.State.NULL);
                video_convert_file.set_state(Gst.State.NULL);
                video_h264_encoder.set_state(Gst.State.NULL);
                file_mp4mux.set_state(Gst.State.NULL);

                audio_src.set_state(Gst.State.NULL);
                audio_filter.set_state(Gst.State.NULL);
                audio_queue_rate.set_state(Gst.State.NULL);
                audio_rate.set_state(Gst.State.NULL);
                audio_convert.set_state(Gst.State.NULL);
                audio_encoder.set_state(Gst.State.NULL);
                audio_queue_mp4mux.set_state(Gst.State.NULL);

                //file_queue.set_state(Gst.State.NULL);
                file_sink.set_state(Gst.State.NULL);

                pipeline.remove_many(video_tee,
                                     video_queue_gtk,
                                     video_queue_file,
                                     video_convert_file,
                                     video_h264_encoder,
                                     audio_src,
                                     audio_filter,
                                     audio_queue_rate,
                                     audio_rate,
                                     audio_convert,
                                     audio_encoder,
                                     audio_queue_mp4mux,
                                     file_mp4mux,
                                     //file_queue,
                                     file_sink);
                var src_pad = video_queue_decoder.get_static_pad("src");
                var sink_pad = video_convert_gtk.get_static_pad("sink");
                src_pad.link(sink_pad);

                video_tee = null;
                video_queue_gtk = null;
                video_queue_file = null;
                video_convert_file = null;
                video_h264_encoder = null;

                /* FIXME: This causes a warning:
                   gst_object_unref:
                   assertion '((GObject *) object)->ref_count > 0' failed
                   Find why. */
                audio_src = null;

                audio_filter = null;
                audio_queue_rate = null;
                audio_rate = null;
                audio_convert = null;
                audio_encoder = null;
                audio_queue_mp4mux = null;
                file_mp4mux = null;
                //file_queue = null;
                file_sink = null;
        }

        /* Closes the file for a H264 saving. */
        private void h264_eos() {
            stderr.printf("h264_eos\n");

            file_mp4mux.set_state(Gst.State.NULL);
            file_sink.set_state(Gst.State.NULL);
            file_sink.set("location", "/dev/null");
            file_sink.set_state(Gst.State.PLAYING);
            file_mp4mux.set_state(Gst.State.PLAYING);

            video_text_record.set("text", "");

            // GLib.Idle.add(() => {
            //         debug();
            //         return GLib.Source.CONTINUE;
            //     });

            //GLib.Idle.add(add_h264_remove_saving_pipeline_block_probe);
            
            //file_queue.set_state(Gst.State.NULL);
            // video_queue_file.set_state(Gst.State.NULL);
            // video_tee.set_state(Gst.State.NULL);

            // audio_queue_mp4mux.set_state(Gst.State.NULL);
            // audio_encoder.set_state(Gst.State.NULL);
            // audio_convert.set_state(Gst.State.NULL);
            // audio_rate.set_state(Gst.State.NULL);
            // audio_queue_rate.set_state(Gst.State.NULL);
            // audio_filter.set_state(Gst.State.NULL);
            // audio_src.set_state(Gst.State.NULL);

            // pipeline.remove_many(file_mp4mux,
            //                      //file_queue,
            //                      file_sink,
            //                      audio_queue_mp4mux,
            //                      audio_encoder,
            //                      audio_convert,
            //                      audio_rate,
            //                      audio_queue_rate,
            //                      audio_filter,
            //                      audio_src,
            //                      video_queue_file,
            //                      video_tee);

            // video_tee = null;
            // video_queue_file = null;

            // /* FIXME: This causes a warning:
            //    gst_object_unref:
            //    assertion '((GObject *) object)->ref_count > 0' failed
            //    Find why. */
            // audio_src = null;

            // audio_filter = null;
            // audio_queue_rate = null;
            // audio_rate = null;
            // audio_convert = null;
            // audio_encoder = null;
            // audio_queue_mp4mux = null;
            // file_mp4mux = null;
            // //file_queue = null;
            // file_sink = null;
        }

        /* Starts the camera asynchronously. */
        private bool start_camera() {
            return change_pipeline_state(Gst.State.PLAYING);
        }

        /* Stops the camera asynchronously. */
        private bool stop_camera() {
            return change_pipeline_state(Gst.State.PAUSED);
        }

        /* Changes the pipeline state, returning until it's changed. */
        private bool change_pipeline_state(Gst.State new_state) {
            pipeline.set_state(new_state);
            Gst.State state = Gst.State.NULL;
            Gst.State pending;

            Gst.StateChangeReturn r;

            do {
                r = pipeline.get_state(out state, out pending, 100);
                if (r == Gst.StateChangeReturn.FAILURE) {
                    return false;
                }
                Thread.usleep(10);
            } while (state != new_state);

            return true;
        }

        /* Sets the raw idle pipeline. */
        private void raw_idle_pipeline() {
            video_src = videodev.create_element("video-src");
            video_filter = Gst.ElementFactory.make("capsfilter",
                                                   "video-filter");
            var fr = video_setup.framerates.first();
            var caps = new Gst.Caps.empty_simple(video_setup.media_type);
            caps.set_simple("framerate", typeof(Gst.Fraction),
                            fr.numerator, fr.denominator,
                            "width", typeof(int), video_setup.width,
                            "height", typeof(int), video_setup.height);
            video_filter.set("caps", caps);
            video_queue_decoder = Gst.ElementFactory.make("queue",
                                                         "video-queue-decoder");
            video_convert_gtk = Gst.ElementFactory.make("videoconvert",
                                                        "video-convert-gtk");
            video_text_record = Gst.ElementFactory.make("textoverlay",
                                                        "video-text-record");
            video_text_record.set("font-desc", "sans 14");
            video_text_record.set("halignment", 2);
            video_text_record.set("valignment", 2);
            video_text_info = Gst.ElementFactory.make("textoverlay",
                                                      "video-text-info");
            video_text_info.set("font-desc", "sans 10");
            video_text_info.set("halignment", 0);
            video_text_info.set("valignment", 2);
            video_text_info.set("line-alignment", 0);
            video_gtk_sink = Gst.ElementFactory.make("gtksink",
                                                     "video-gtk-sink");

            pipeline.add_many(video_src,
                              video_filter,
                              video_queue_decoder,
                              video_convert_gtk,
                              video_text_record,
                              video_text_info,
                              video_gtk_sink);
            video_src.link_many(video_filter,
                                video_queue_decoder,
                                video_convert_gtk,
                                video_text_record,
                                video_text_info,
                                video_gtk_sink);
        }

        /* Sets the H264 idle pipeline. */
        private void h264_idle_pipeline() {
            stderr.printf("%s\n", video_setup.to_string());
            video_src = videodev.create_element("video-src");
            video_filter = Gst.ElementFactory.make("capsfilter",
                                                   "video-filter");
            var fr = video_setup.framerates.first();
            var caps = new Gst.Caps.empty_simple(video_setup.media_type);
            caps.set_simple("framerate", typeof(Gst.Fraction),
                            fr.numerator, fr.denominator,
                            "width", typeof(int), video_setup.width,
                            "height", typeof(int), video_setup.height);
            video_filter.set("caps", caps);
            video_queue_parse = Gst.ElementFactory.make("queue",
                                                        "video-queue-parse");
            video_queue_parse.set("max-size-time", 3 * Gst.SECOND);
            video_queue_parse.set("max-size-bytes", 0);
            video_queue_parse.set("max-size-buffers", 0);
            video_queue_parse.set("leaky", 1);
            video_queue_parse.set("flush-on-eos", true);
            video_h264_parse = Gst.ElementFactory.make("h264parse",
                                                       "video-h264-parse");
            video_h264_parse.set("config-interval", -1);
            video_h264_parse.set("disable-passthrough", true);
            video_queue_gtk = Gst.ElementFactory.make("queue",
                                                      "video-queue-gtk");
            video_h264_decoder = Gst.ElementFactory.make("avdec_h264",
                                                         "video-h264-decoder");
            video_queue_convert = Gst.ElementFactory.make("queue",
                                                          "video-queue-convert");
            video_convert_gtk = Gst.ElementFactory.make("videoconvert",
                                                        "video-convert-gtk");
            video_text_record = Gst.ElementFactory.make("textoverlay",
                                                        "video-text-record");
            video_text_record.set("font-desc", "sans 14");
            video_text_record.set("halignment", 2);
            video_text_record.set("valignment", 2);
            video_text_info = Gst.ElementFactory.make("textoverlay",
                                                      "video-text-info");
            video_text_info.set("font-desc", "sans 10");
            video_text_info.set("halignment", 0);
            video_text_info.set("valignment", 2);
            video_text_info.set("line-alignment", 0);
            video_gtk_sink = Gst.ElementFactory.make("gtksink",
                                                     "video-gtk-sink");

            pipeline.add_many(video_src,
                              video_filter,
                              video_queue_parse,
                              video_h264_parse,
                              video_queue_gtk,
                              video_h264_decoder,
                              video_queue_convert,
                              video_convert_gtk,
                              video_text_record,
                              video_text_info,
                              video_gtk_sink);
            video_src.link_many(video_filter,
                                video_queue_parse,
                                video_h264_parse,
                                video_queue_gtk,
                                video_h264_decoder,
                                video_queue_convert,
                                video_convert_gtk,
                                video_text_record,
                                video_text_info,
                                video_gtk_sink);
        }

        /* Starts saving. */
        private void start_saving() {
            stderr.printf("start_saving\n");
            var pad = video_queue_parse.get_static_pad("src");
            pad.add_probe(Gst.PadProbeType.BLOCK | Gst.PadProbeType.BUFFER,
                          start_saving_block_probe);
        }

        private bool real_start_saving() {
            Thread.usleep(100);
            stderr.printf("real_start_saving\n");
            start_saving_ts = last_saving_ts = Timestamp.get_timestamp();
            video_text_record.set("text", "•REC 00:00:00");
            video_text_record.set("halignment", 2);
            video_text_record.set("valignment", 2);
            GLib.Idle.add(update_counter);
            buffer_count = 0;
            drop_count = 0;
            var pad = video_queue_file.get_static_pad("src");
            pad.add_probe(Gst.PadProbeType.BUFFER, second_drop_one_probe);
            stderr.printf("REMOVING [%lu]\n", block_probe_id);
            pad.remove_probe(block_probe_id);
            block_probe_id = 0;
            return GLib.Source.REMOVE;
        }

        /* Stops saving. */
        private void stop_saving() {
            stderr.printf("stop_saving\n");
            var pad = video_queue_parse.get_static_pad("src");
            pad.add_probe(Gst.PadProbeType.BLOCK_DOWNSTREAM,
                          stop_saving_block_probe);
        }

        /* Change video source block probe. */
        private Gst.PadProbeReturn
        start_saving_block_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            stderr.printf("start_saving_block_probe %s::%s\n",
                          pad.parent.name, pad.name);
            pad.remove_probe(info.id);

            switch (video_setup.media_type) {
            case VideoSetup.RAW:
                raw_add_saving_pipeline();
                break;
            case VideoSetup.H264:
                h264_add_saving_pipeline();
                break;
            }

            /*new Thread<void*>("add-wait-keyframe-thread",
              add_wait_keyframe_probe);*/

            return Gst.PadProbeReturn.DROP;
        }

        private void raw_add_saving_pipeline() {
            var src_pad = video_h264_parse.get_static_pad("src");
            var sink_pad = video_queue_gtk.get_static_pad("sink");

            src_pad.unlink(sink_pad);

            audio_src = audiodev.create_element("audio-src");
            audio_src.set("stream-id", "Audio stream");
            audio_filter = Gst.ElementFactory.make("capsfilter",
                                                   "audio-filter");
            int rate = get_audio_rate();
            int channels = audio_setup.max_rate > 1 ? 2 : 1;
            var fmt = audio_setup.formats.first();
            var acaps = new Gst.Caps.empty_simple(audio_setup.media_type);
            acaps.set_simple("format", typeof(string), fmt.format,
                             "rate", typeof(int), rate,
                             "channels", typeof(int), channels);
            audio_filter.set("caps", acaps);
            audio_queue_rate = Gst.ElementFactory.make("queue",
                                                       "audio-queue-rate");
            audio_rate = Gst.ElementFactory.make("audiorate", "audio-rate");
            audio_convert = Gst.ElementFactory.make("audioconvert",
                                                    "audio-convert");
            audio_encoder = Gst.ElementFactory.make("avenc_aac",
                                                    "audio_encoder");
            audio_queue_mp4mux = Gst.ElementFactory.make("queue",
                                                         "audio_queue_mp4mux");
            audio_queue_mp4mux.set("max-size-time", 3 * Gst.SECOND);
            audio_queue_mp4mux.set("max-size-bytes", 0);
            audio_queue_mp4mux.set("max-size-buffers", 0);
            audio_queue_mp4mux.set("leaky", 1);

            video_tee = Gst.ElementFactory.make("tee", "video-tee");
            video_queue_file = Gst.ElementFactory.make("queue",
                                                       "video-queue-file");
            // video_queue_file.set("max-size-time", Gst.SECOND);
            // video_queue_file.set("max-size-bytes", 0);
            // video_queue_file.set("max-size-buffers", 0);
            // video_queue_file.set("leaky", 1);
            video_convert_file = Gst.ElementFactory.make("videoconvert",
                                                         "video-convert-file");
            video_h264_encoder = Gst.ElementFactory.make("x264enc",
                                                         "video-h264-encoder");
            file_mp4mux = Gst.ElementFactory.make("mpegtsmux",
                                                  "file-mp4mux");
            //file_queue = Gst.ElementFactory.make("queue", "file-queue");
            file_sink = Gst.ElementFactory.make("filesink", "file-sink");
            //file_sink.set("sync", true);
            file_sink.set("async", false);
            file_sink.set("location", temporary_fn);

            pipeline.add_many(video_tee,
                              audio_src,
                              audio_filter,
                              audio_queue_rate,
                              audio_rate,
                              audio_convert,
                              audio_encoder,
                              audio_queue_mp4mux,
                              video_queue_gtk,
                              video_queue_file,
                              video_convert_file,
                              video_h264_encoder,
                              file_mp4mux,
                              //file_queue,
                              file_sink);
            video_queue_decoder.link(video_tee);
            video_tee.link_many(video_queue_gtk,
                                video_convert_gtk);
            video_tee.link_many(video_queue_file,
                                video_convert_file,
                                video_h264_encoder,
                                file_mp4mux,
                                //file_queue,
                                file_sink);
            audio_src.link_many(audio_filter,
                                audio_queue_rate,
                                audio_rate,
                                audio_convert,
                                audio_encoder,
                                audio_queue_mp4mux,
                                file_mp4mux);
        }

        private void h264_add_saving_pipeline() {
            stderr.printf("h264_add_saving_pipeline\n");
            var src_pad = video_h264_parse.get_static_pad("src");
            var sink_pad = video_queue_gtk.get_static_pad("sink");
            src_pad.unlink(sink_pad);

            audio_src = audiodev.create_element("audio-src");
            audio_filter = Gst.ElementFactory.make("capsfilter",
                                                   "audio-filter");
            int rate = get_audio_rate();
            int channels = audio_setup.max_rate > 1 ? 2 : 1;
            var fmt = audio_setup.formats.first();
            var acaps = new Gst.Caps.empty_simple(audio_setup.media_type);
            acaps.set_simple("format", typeof(string), fmt.format,
                             "rate", typeof(int), rate,
                             "channels", typeof(int), channels);
            audio_filter.set("caps", acaps);
            audio_queue_rate = Gst.ElementFactory.make("queue",
                                                       "audio-queue-rate");
            audio_rate = Gst.ElementFactory.make("audiorate", "audio-rate");
            audio_convert = Gst.ElementFactory.make("audioconvert",
                                                    "audio-convert");
            audio_encoder = Gst.ElementFactory.make("avenc_aac",
                                                    "audio_encoder");
            audio_queue_mp4mux = Gst.ElementFactory.make("queue",
                                                         "audio_queue_mp4mux");
            audio_queue_mp4mux.set("max-size-time", 3 * Gst.SECOND);
            audio_queue_mp4mux.set("max-size-bytes", 0);
            audio_queue_mp4mux.set("max-size-buffers", 0);
            audio_queue_rate.set("leaky", 1);
            audio_queue_rate.set("flush-on-eos", true);

            video_tee = Gst.ElementFactory.make("tee", "video-tee");
            video_queue_file = Gst.ElementFactory.make("queue",
                                                       "video-queue-file");
            video_queue_file.set("max-size-time", 3 * Gst.SECOND);
            video_queue_file.set("max-size-bytes", 0);
            video_queue_file.set("max-size-buffers", 0);
            video_queue_file.set("leaky", 1);
            video_queue_file.set("flush-on-eos", true);
            file_mp4mux = Gst.ElementFactory.make("mpegtsmux",
                                                  "file-mp4mux");
            // file_mp4mux.set("alignment", 0);
            // file_mp4mux.set("pat-interval", 100);
            // file_mp4mux.set("pmt-interval", 100);
            // file_mp4mux.set("si-interval", 100);
            file_sink = Gst.ElementFactory.make("filesink", "file-sink");
            file_sink.set("sync", true);
            file_sink.set("async", false);
            file_sink.set("location", "/dev/null");

            pipeline.add_many(audio_src,
                              audio_filter,
                              audio_queue_rate,
                              audio_rate,
                              audio_convert,
                              audio_encoder,
                              audio_queue_mp4mux,
                              video_tee,
                              video_queue_file,
                              file_mp4mux,
                              file_sink);
            video_h264_parse.link(video_tee);
            video_tee.link(video_queue_gtk);
            video_tee.link_many(video_queue_file,
                                file_mp4mux,
                                file_sink);

            audio_src.link_many(audio_filter,
                                audio_queue_rate,
                                audio_rate,
                                audio_convert,
                                audio_encoder,
                                audio_queue_mp4mux,
                                file_mp4mux);

            audio_src.set_state(Gst.State.PLAYING);
            audio_filter.set_state(Gst.State.PLAYING);
            audio_queue_rate.set_state(Gst.State.PLAYING);
            audio_rate.set_state(Gst.State.PLAYING);
            audio_convert.set_state(Gst.State.PLAYING);
            audio_encoder.set_state(Gst.State.PLAYING);
            audio_queue_mp4mux.set_state(Gst.State.PLAYING);
            video_tee.set_state(Gst.State.PLAYING);
            video_queue_file.set_state(Gst.State.PLAYING);
            file_mp4mux.set_state(Gst.State.PLAYING);
            file_sink.set_state(Gst.State.PLAYING);

            GLib.Idle.add(h264_add_first_block_probe);
        }

        private int buffer_count;
        private ulong block_probe_id;

        private bool h264_add_first_block_probe() {
            stderr.printf("h264_add_first_block_probe\n");
            Thread.usleep(1000);
            var pad = video_queue_file.get_static_pad("src");
            block_probe_id = pad.add_probe(Gst.PadProbeType.BLOCK |
                                           Gst.PadProbeType.BUFFER,
                                           h264_first_block_probe);
            return GLib.Source.REMOVE;
        }

        private Gst.PadProbeReturn
        h264_first_block_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            stderr.printf("h264_first_block_probe %s::%s blocked (1)\n",
                          pad.parent.name, pad.name);
            GLib.Idle.add(h264_first_unblock_probe);
            return Gst.PadProbeReturn.OK;
        }

        private bool h264_first_unblock_probe() {
            stderr.printf("h264_first_unblock_probe\n");
            Thread.usleep(100);
            var pad = video_queue_file.get_static_pad("src");
            buffer_count = 0;
            drop_count = 0;
            pad.add_probe(Gst.PadProbeType.BUFFER, first_drop_one_probe);
            pad.remove_probe(block_probe_id);
            block_probe_id = 0;
            return GLib.Source.REMOVE;
        }

        private int drop_count;

        private Gst.PadProbeReturn
        first_drop_one_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            stderr.printf("first_drop_one_probe %s::%s\n",
                          pad.parent.name, pad.name);
            GLib.Idle.add(h264_first_stop_probe);
            return Gst.PadProbeReturn.REMOVE;
            // if (buffer_count++ == 0) {
            //     stderr.printf("Drop one buffer...\n");
            //     return Gst.PadProbeReturn.DROP;
            // }
            // Gst.Buffer buffer = (Gst.Buffer)info.data;
            // var flags = buffer.get_flags();
            // if ((flags & Gst.BufferFlags.DELTA_UNIT) == 0) {
            //     stderr.printf("keyframe found!\n");
            //     GLib.Idle.add(h264_first_stop_probe);
            //     return Gst.PadProbeReturn.REMOVE;
            // } else {
            //     stderr.printf("waiting 1 keyframe %d...\n", drop_count++);
            //     return Gst.PadProbeReturn.DROP;
            // }
        }

        private Gst.PadProbeReturn
        second_drop_one_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            stderr.printf("second_drop_one_probe %s::%s\n",
                          pad.parent.name, pad.name);
            return Gst.PadProbeReturn.REMOVE;
            // if (buffer_count++ == 0) {
            //     stderr.printf("Drop one buffer...\n");
            //     return Gst.PadProbeReturn.DROP;
            // }
            // Gst.Buffer buffer = (Gst.Buffer)info.data;
            // var flags = buffer.get_flags();
            // if ((flags & Gst.BufferFlags.DELTA_UNIT) == 0) {
            //     stderr.printf("keyframe found!\n");
            //     pad.remove_probe(info.id);
            //     debug();
            //     return Gst.PadProbeReturn.REMOVE;
            // } else {
            //     stderr.printf("waiting 2 keyframe %d...\n", drop_count++);
            //     return Gst.PadProbeReturn.DROP;
            // }
        }

        private bool first_eos;

        private bool h264_first_stop_probe() {
            stderr.printf("h264_first_stop_probe\n");
            Thread.usleep(100);
            var pad = video_queue_file.get_static_pad("src");
            block_probe_id = pad.add_probe(Gst.PadProbeType.BLOCK |
                                           Gst.PadProbeType.BUFFER,
                                           h264_second_block_probe);
            stderr.printf("ADDED [%lu]\n", block_probe_id);

            first_eos = true;
            new Thread<void*>("first-push-eos-thread", first_push_eos_thread);

            return GLib.Source.REMOVE;
        }

        private void* first_push_eos_thread() {
            stderr.printf("first_push_eos_thread\n");
            int c = 0;
            while (!block_probe_called) {
                stderr.printf("Waiting for probe %d....\n", c++);
                Thread.usleep(1000);
            }
            var pad = file_mp4mux.get_static_pad("src");
            var peer = pad.get_peer();
            pipeline.message_forward = true;
            peer.send_event(new Gst.Event.eos());
            return null;
        }

        private bool block_probe_called = false;

        private Gst.PadProbeReturn
        h264_second_block_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            stderr.printf("h264_second_stop_probe %s::%s [%lu] blocked (2)\n",
                          pad.parent.name, pad.name, info.id);
            block_probe_called = true;
            return Gst.PadProbeReturn.OK;
        }

        /* Saving block probe. */
        private Gst.PadProbeReturn
        stop_saving_block_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            // stderr.printf("stop_saving_block_probe\n");
            pad.remove_probe(info.id);
            // switch (video_setup.media_type) {
            // case VideoSetup.RAW:
            //     //raw_remove_saving_pipeline();
            //     break;
            // case VideoSetup.H264:
            //     //h264_remove_saving_pipeline();
            //     break;
            // }

            var peer = pad.get_peer();
            pipeline.message_forward = true;
            peer.send_event(new Gst.Event.eos());

            //new Thread<void*>("eos-push-thread", push_eos_thread);

            return Gst.PadProbeReturn.DROP;
        }

        private void raw_remove_saving_pipeline() {
            var src = video_queue_decoder.get_static_pad("src");
            var sink = video_tee.get_static_pad("sink");
            src.unlink(sink);
            src = video_queue_gtk.get_static_pad("src");
            sink = video_convert_gtk.get_static_pad("sink");
            src.unlink(sink);
            src = video_queue_decoder.get_static_pad("src");
            sink = video_convert_gtk.get_static_pad("sink");
            src.link(sink);
        }

        private Gst.PadProbeReturn
        h264_remove_saving_pipeline_block_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            stderr.printf("h264_remove_saving_pipeline_block_probe %s::%s\n",
                          pad.parent.name, pad.name);
            pad.remove_probe(info.id);
            video_h264_parse.unlink(video_tee);
            video_tee.unlink(video_queue_gtk);
            audio_queue_mp4mux.unlink(file_mp4mux);
            video_h264_parse.link(video_queue_gtk);
            video_queue_file.set_state(Gst.State.NULL);
            video_tee.set_state(Gst.State.NULL);
            audio_queue_mp4mux.set_state(Gst.State.NULL);
            audio_encoder.set_state(Gst.State.NULL);
            audio_convert.set_state(Gst.State.NULL);
            audio_rate.set_state(Gst.State.NULL);
            audio_queue_rate.set_state(Gst.State.NULL);
            audio_filter.set_state(Gst.State.NULL);
            audio_src.set_state(Gst.State.NULL);

            pipeline.remove_many(file_mp4mux,
                                 file_sink,
                                 audio_queue_mp4mux,
                                 audio_encoder,
                                 audio_convert,
                                 audio_rate,
                                 audio_queue_rate,
                                 audio_filter,
                                 audio_src,
                                 video_queue_file,
                                 video_tee);

            video_tee = null;
            video_queue_file = null;

            /* FIXME: This causes a warning:
               gst_object_unref:
               assertion '((GObject *) object)->ref_count > 0' failed
               Find why. */
            audio_src = null;

            audio_filter = null;
            audio_queue_rate = null;
            audio_rate = null;
            audio_convert = null;
            audio_encoder = null;
            audio_queue_mp4mux = null;
            file_mp4mux = null;
            file_sink = null;

            return Gst.PadProbeReturn.DROP;
        }

        private bool
        add_h264_remove_saving_pipeline_block_probe() {
            stderr.printf("add_h264_remove_saving_pipeline_block_probe\n");
            var pad = video_queue_file.get_static_pad("src");
            pad.add_probe(Gst.PadProbeType.BLOCK | Gst.PadProbeType.BUFFER,
                          h264_remove_saving_pipeline_block_probe);
            return GLib.Source.REMOVE;
        }

        private void device_changed(DeviceType device_type) {
            if (saving)
                return;
            switch (device_type) {
            case VIDEO:
                video_device_changed();
                break;
            }
        }

        private Gst.Element new_video_src;
        private Gst.Element new_video_filter;
        private Gst.Element new_video_queue_decoder;

        private void video_device_changed() {
            if (!last_change_done)
                return;
            last_change_done = false;
            new_videodev = devices.video_device;
            new_video_setup = new_videodev.setups.first();
            if (video_setup.media_type == VideoSetup.RAW &&
                new_video_setup.media_type == VideoSetup.H264)
                stderr.printf("unimplemented\n");
            if (video_setup.media_type == VideoSetup.H264 &&
                new_video_setup.media_type == VideoSetup.RAW)
                h2642raw_block_video_source();
        }

        private void raw_block_video_source() {
            var pad = video_queue_decoder.get_static_pad("src");
            pad.add_probe(Gst.PadProbeType.BLOCK |
                          Gst.PadProbeType.BUFFER,
                          change_video_source_block_probe);
        }

        private void h2642raw_block_video_source() {
            var pad = video_queue_gtk.get_static_pad("src");
            pad.add_probe(Gst.PadProbeType.BLOCK |
                          Gst.PadProbeType.BUFFER,
                          h2642raw_block_probe);
        }

        private Gst.PadProbeReturn
        h2642raw_block_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            pad.remove_probe(info.id);
            video_filter.set_state(Gst.State.NULL);
            video_queue_parse.set_state(Gst.State.NULL);
            video_h264_parse.set_state(Gst.State.NULL);
            video_queue_decoder.set_state(Gst.State.NULL);
            video_h264_decoder.set_state(Gst.State.NULL);

            video_h264_decoder.unlink(video_queue_gtk);

            pipeline.remove_many(video_filter,
                                 video_queue_parse,
                                 video_h264_parse,
                                 video_queue_decoder,
                                 video_h264_decoder);
            new Thread<void*>("cut_h264_video_src", h2642raw_thread);
            return Gst.PadProbeReturn.DROP;
        }

        private void*
        h2642raw_thread() {
            change_pipeline_state(Gst.State.NULL);
            video_src.set_state(Gst.State.NULL);
            pipeline.remove(video_src);

            videodev = new_videodev;
            video_setup = videodev.setups.first();
            new_videodev = null;

            new_video_src = videodev.create_element("video-src");
            new_video_filter = Gst.ElementFactory.make("capsfilter",
                                                   "video-filter");
            var fr = video_setup.framerates.first();
            var caps = new Gst.Caps.empty_simple(video_setup.media_type);
            caps.set_simple("framerate", typeof(Gst.Fraction),
                            fr.numerator, fr.denominator,
                            "width", typeof(int), video_setup.width,
                            "height", typeof(int), video_setup.height);
            new_video_filter.set("caps", caps);
            pipeline.add_many(new_video_src,
                              new_video_filter);
            new_video_src.link_many(new_video_filter,
                                    video_queue_gtk);

            new_video_src.set_state(Gst.State.PLAYING);
            new_video_filter.set_state(Gst.State.PLAYING);
            change_pipeline_state(Gst.State.PLAYING);

            new Thread<void*>("update_references", update_references);

            return null;
        }

        private void* update_references() {
            Thread.usleep(10000000);
            video_filter.set_state(Gst.State.NULL);
            video_queue_parse.set_state(Gst.State.NULL);
            video_h264_parse.set_state(Gst.State.NULL);
            video_queue_decoder.set_state(Gst.State.NULL);
            video_h264_decoder.set_state(Gst.State.NULL);

            video_queue_parse = video_h264_parse = video_queue_decoder
            = video_h264_decoder = null;

            video_src = new_video_src;
            video_filter = new_video_filter;
            last_change_done = true;
            return null;
        }

        /* Change video source block probe. */
        private Gst.PadProbeReturn
        change_video_source_block_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            stderr.printf("change_video_source_block_probe\n");
            pad.remove_probe(info.id);

            switch (video_setup.media_type) {
            case VideoSetup.RAW:
                raw_change_video_source();
                break;
            case VideoSetup.H264:
                h264_change_video_source();
                break;
            }

            new Thread<void*>("add-wait-keyframe-thread",
                              add_wait_keyframe_probe);

            return Gst.PadProbeReturn.DROP;
        }

        private void raw_change_video_source() {
            videodev = new_videodev;
            video_setup = videodev.setups.first();
            new_videodev = null;

            switch (video_setup.media_type) {
            case VideoSetup.RAW:
                raw_raw_change_video_source();
                break;
            case VideoSetup.H264:
                raw_h264_change_video_source();
                break;
            }
        }

        private void h264_change_video_source() {
            stderr.printf("h264_change_video_source\n");
            videodev = new_videodev;
            new_videodev = null;
            video_setup = videodev.setups.first();

            switch (video_setup.media_type) {
            case VideoSetup.RAW:
                h264_raw_change_video_source();
                break;
            case VideoSetup.H264:
                h264_h264_change_video_source();
                break;
            }
        }

        private void raw_raw_change_video_source() {
            video_src.set_state(Gst.State.NULL);
            video_filter.set_state(Gst.State.NULL);
            pipeline.remove_many(video_src,
                                 video_filter);

            video_src = videodev.create_element("video-src");
            video_filter = Gst.ElementFactory.make("capsfilter",
                                                   "video-filter");
            var fr = video_setup.framerates.first();
            var caps = new Gst.Caps.empty_simple(video_setup.media_type);
            caps.set_simple("framerate", typeof(Gst.Fraction),
                            fr.numerator, fr.denominator,
                            "width", typeof(int), video_setup.width,
                            "height", typeof(int), video_setup.height);
            video_filter.set("caps", caps);
            pipeline.add_many(video_src,
                              video_filter);
            video_src.link_many(video_filter,
                                video_queue_decoder);
            video_src.set_state(Gst.State.PLAYING);
            video_filter.set_state(Gst.State.PLAYING);
        }

        private void raw_h264_change_video_source() {
            video_src.set_state(Gst.State.NULL);
            video_filter.set_state(Gst.State.NULL);
            pipeline.remove_many(video_src,
                                 video_filter);

            video_src  = videodev.create_element("video-src");
            video_filter = Gst.ElementFactory.make("capsfilter",
                                                        "video-filter");
            var fr = video_setup.framerates.first();
            var caps = new Gst.Caps.empty_simple(video_setup.media_type);
            caps.set_simple("framerate", typeof(Gst.Fraction),
                            fr.numerator, fr.denominator,
                            "width", typeof(int), video_setup.width,
                            "height", typeof(int), video_setup.height);
            video_filter.set("caps", caps);
            video_queue_parse = Gst.ElementFactory.make("queue",
                                                        "video-queue-parse");
            video_h264_parse = Gst.ElementFactory.make("h264parse",
                                                       "video-h264-parse");
            pipeline.add_many(video_src,
                              video_filter,
                              video_queue_parse,
                              video_h264_parse);
            video_src.link_many(video_filter,
                                video_queue_parse,
                                video_h264_parse,
                                video_queue_decoder);
            video_src.set_state(Gst.State.PLAYING);
            video_filter.set_state(Gst.State.PLAYING);
            video_queue_parse.set_state(Gst.State.PLAYING);
            video_h264_parse.set_state(Gst.State.PLAYING);
        }

        private Gst.State get_state(Gst.Element element) {
            Gst.State state;
            Gst.State pending;
            Gst.StateChangeReturn r;
            r = pipeline.get_state(out state, out pending, 100);
            if (r == Gst.StateChangeReturn.FAILURE) {
                stderr.printf("%s: WTF!\n", element.name);
                return state;
            }
            return state;
        }

        private void h264_raw_change_video_source() {
            stderr.printf("h264_raw_change_video_source\n");
            video_src.set_state(Gst.State.NULL);
            video_filter.set_state(Gst.State.NULL);
            video_queue_parse.set_state(Gst.State.NULL);
            video_h264_parse.set_state(Gst.State.NULL);

            pipeline.remove_many(video_src,
                                 video_filter,
                                 video_queue_parse,
                                 video_h264_parse);

            video_queue_parse = video_h264_parse = null;

            video_src = videodev.create_element("video-src");
            video_filter = Gst.ElementFactory.make("capsfilter",
                                                   "video-filter");
            var fr = video_setup.framerates.first();
            var caps = new Gst.Caps.empty_simple(video_setup.media_type);
            caps.set_simple("framerate", typeof(Gst.Fraction),
                            fr.numerator, fr.denominator,
                            "width", typeof(int), video_setup.width,
                            "height", typeof(int), video_setup.height);
            video_filter.set("caps", caps);
            pipeline.add_many(video_src,
                              video_filter);
            video_src.link_many(video_filter,
                                video_queue_decoder);

            video_src.set_state(Gst.State.PLAYING);
            video_filter.set_state(Gst.State.PLAYING);
        }

        private void h264_h264_change_video_source() {
            video_src.set_state(Gst.State.NULL);
            video_filter.set_state(Gst.State.NULL);
            pipeline.remove_many(video_src, video_filter);

            video_src = videodev.create_element("video-src");
            video_filter = Gst.ElementFactory.make("capsfilter",
                                                   "video-filter");
            var fr = video_setup.framerates.first();
            var caps = new Gst.Caps.empty_simple(video_setup.media_type);
            caps.set_simple("framerate", typeof(Gst.Fraction),
                            fr.numerator, fr.denominator,
                            "width", typeof(int), video_setup.width,
                            "height", typeof(int), video_setup.height);
            video_filter.set("caps", caps);
            pipeline.add_many(video_src,
                              video_filter);
            // video_src.link_many(video_filter,
            //                     video_queue_parse);
            video_src.set_state(Gst.State.PLAYING);
            video_filter.set_state(Gst.State.PLAYING);
        }

        private void*
        add_wait_keyframe_probe() {
            var pad = video_queue_file.get_static_pad("src");
            pad.add_probe(Gst.PadProbeType.BLOCK | Gst.PadProbeType.BUFFER,
                          wait_keyframe_probe);
            return null;
        }

        private Gst.PadProbeReturn
        wait_keyframe_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            Gst.Buffer buffer = (Gst.Buffer)info.data;
            var flags = buffer.get_flags();
            if ((flags & Gst.BufferFlags.DELTA_UNIT) == 0) {
                return Gst.PadProbeReturn.REMOVE;
            } else {
                return Gst.PadProbeReturn.DROP;
            }
        }

        /* Gets the final destination for the video. */
        private string get_destination() {
            var now = new GLib.DateTime.now_local();
            string sts = "%04d-%02d-%02d %02d:%02d:%02d".printf(
                now.get_year(), now.get_month(), now.get_day_of_month(),
                now.get_hour(), now.get_minute(), now.get_second());
            var video_dir = GLib.Environment.get_user_special_dir(
                GLib.UserDirectory.VIDEOS);
            return string.join(GLib.Path.DIR_SEPARATOR_S, video_dir,
                               "Queso-%s.mp4".printf(sts));
        }

        /* Pushes the EOS event. */
        private void* push_eos_thread() {
            stderr.printf("push_eos_thread\n");
            var pad = file_mp4mux.get_static_pad("src");
            var peer = pad.get_peer();
            pipeline.message_forward = true;
            peer.send_event(new Gst.Event.eos());
            return null;
        }

        /* Moves a file. */
        private void move_file(string src_path, string dest_path) {
            GLib.File src = GLib.File.new_for_path(src_path);
            GLib.File dest = GLib.File.new_for_path(dest_path);

            try {
                src.move(dest, GLib.FileCopyFlags.NONE, null);
            } catch (GLib.Error e) {
                GLib.warning("Could not move “%s” to “%s”: %s",
                             src_path, dest_path, e.message);
            }
        }

        /* Update counter idle callback. */
        private bool update_counter() {
            if (!saving) {
                video_text_record.set("text", "");
                return GLib.Source.REMOVE;
            }
            Gst.State state = Gst.State.NULL;
            Gst.State pending;
            pipeline.get_state(out state, out pending, 100);
            if (state != Gst.State.PLAYING)
                return GLib.Source.CONTINUE;
            double ts = Timestamp.get_timestamp();
            if (ts - last_saving_ts < 1.0)
                return GLib.Source.CONTINUE;
            last_saving_ts = ts;
            string sts = Timestamp.to_string(ts - start_saving_ts);
            video_text_record.set("text", "•REC " + sts);
            return GLib.Source.CONTINUE;
        }

        /* Gets a random filename. */
        private string random_filename() {
            int ts = (int)Timestamp.get_timestamp();
            return string.join(GLib.Path.DIR_SEPARATOR_S,
                               GLib.Environment.get_tmp_dir(),
                               "queso-%08X.mp4".printf(ts));
        }

        private int get_audio_rate() {
            int rate = DEFAULT_AUDIO_RATE;
            while (rate > audio_setup.max_rate)
                rate /= 2;
            return rate;
        }

        private string get_info() {
            var r = new StringBuilder();
            r.append("Video source: ");
            r.append(videodev.name);
            r.append("\n");
            r.append("Audio source: ");
            r.append(audiodev.name);
            return r.str;
        }

        private void try_link(Gst.Element src, Gst.Element sink) {
            if (!src.link(sink)) {
                stderr.printf("%s -> %s failed!\n",
                              src.name, sink.name);
                debug();
            }
        }

        private void link_pads(Gst.Element src, Gst.Element sink) {
            string a = src.name;
            string b = sink.name;
            var src_pad = src.get_static_pad("src");
            var sink_pad = sink.get_static_pad("sink");
            switch (src_pad.link(sink_pad)) {
            case OK:
                stderr.printf("%s -> %s: %s!\n", a, b, "OK");
                break;
            case WRONG_HIERARCHY:
                stderr.printf("%s -> %s: %s!\n", a, b, "WRONG_HIERARCHY");
                break;
            case WAS_LINKED:
                stderr.printf("%s -> %s: %s!\n", a, b, "WAS_LINKED");
                break;
            case WRONG_DIRECTION:
                stderr.printf("%s -> %s: %s!\n", a, b, "WRONG_DIRECTION");
                break;
            case NOFORMAT:
                stderr.printf("%s -> %s: %s!\n", a, b, "NOFORMAT");
                break;
            case NOSCHED:
                stderr.printf("%s -> %s: %s!\n", a, b, "NOSCHED");
                break;
            }
        }
    }
}
