/*
 * This file is part of queso.
 *
 * Copyright © 2020 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace Queso {

    /**
     * Class to manage devices in the system.
     */
    public class Devices : GLib.Object {

        public signal void device_changed(DeviceType device_type);

        /**
         * Null device.
         */
        public const string DEV_NULL = "/dev/null";

        /**
         * The video devices.
         */
        public Gee.TreeSet<VideoDevice> video_devices { public get; private set; }

        /**
         * The audio devices.
         */
        public Gee.TreeSet<AudioDevice> audio_devices { public get; private set; }

        /**
         * The selected video device.
         */
        public VideoDevice video_device { public get; private set; }

        /**
         * The selected audio device.
         */
        public AudioDevice audio_device { public get; private set; }

        private Gee.Iterator<VideoDevice> vd_iter;

        /* The device monitor. */
        private Gst.DeviceMonitor monitor;
        /* The bus. */
        private Gst.Bus bus;

        /* Constructs a device manager. */
        construct {
            audio_devices = new Gee.TreeSet<AudioDevice>();
            video_devices = new Gee.TreeSet<VideoDevice>();
            audio_devices.add(new AudioDevice.test());
            video_devices.add(new VideoDevice.test());
            monitor = new Gst.DeviceMonitor();
            bus = monitor.get_bus();
            bus.add_watch(GLib.Priority.DEFAULT, monitor_devices);

            monitor.add_filter("Audio/Source",
                               new Gst.Caps.empty_simple(AudioSetup.RAW));
            monitor.add_filter("Video/Source",
                               new Gst.Caps.empty_simple(VideoSetup.RAW));
            monitor.add_filter("Video/Source",
                               new Gst.Caps.empty_simple(VideoSetup.H264));

            monitor.start();

            var devices = monitor.get_devices();
            unowned GLib.List<Gst.Device> t = devices;
            while (t != null) {
                switch (t.data.get_device_class()) {
                case "Audio/Source":
                    if (!("Monitor of" in t.data.get_display_name()))
                        audio_devices.add(new AudioDevice(t.data));
                    break;
                case "Video/Source":
                    video_devices.add(new VideoDevice(t.data));
                    break;
                }
                t = t.next;
            }

            vd_iter = video_devices.iterator();
            if (vd_iter.has_next()) {
                vd_iter.next();
                video_device = vd_iter.get();
            }
            audio_device = audio_devices.first();
        }

        public void next_video_device() {
            if (!vd_iter.has_next())
                vd_iter = video_devices.iterator();
            vd_iter.next();
            video_device = vd_iter.get();
            device_changed(DeviceType.VIDEO);
        }

        public void update_video_device(string name) {
            if (video_device.name == name)
                return;
            foreach (var device in video_devices) {
                if (device.name == name) {
                    video_device = device;
                    device_changed(DeviceType.VIDEO);
                }
            }
        }

        public void update_audio_device(string name) {
            if (video_device.name == name)
                return;
            foreach (var device in audio_devices) {
                if (device.name == name) {
                    audio_device = device;
                    device_changed(DeviceType.AUDIO);
                }
            }
        }

        /* Monitors the devices. */
        private bool monitor_devices(Gst.Bus bus, Gst.Message message) {
            Gst.Device device;
            string name;

            switch (message.type) {
            case Gst.MessageType.DEVICE_ADDED:
                message.parse_device_added(out device);
                name = device.get_display_name();
                stdout.printf("Device added: %s\n", name);
                break;
            case Gst.MessageType.DEVICE_REMOVED:
                message.parse_device_removed(out device);
                name = device.get_display_name();
                stdout.printf("Device added: %s\n", name);
                break;
            default:
                break;
            }

            return GLib.Source.CONTINUE;
        }
    }
}
