/*
 * This file is part of queso.
 *
 * Copyright © 2020 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace Queso {

    /**
     * Class for video devices.
     */
    public class VideoDevice : Device, Gee.Comparable<VideoDevice> {

        public string device_path { public get; private set; }

        /**
         * The device setups.
         */
        public Gee.TreeSet<VideoSetup> setups { public get; private set; }

        /**
         * Constructs a video device.
         * @param device the GStreamer device.
         */
        public VideoDevice(Gst.Device device) {
            base(device);
            var props = device.get_properties();
            device_path = Devices.DEV_NULL;
            if (!props.has_field("device.path"))
                return;
            unowned GLib.Value v = props.get_value("device.path");
            if (v.type() != typeof(string))
                return;
            device_path = v.dup_string();
        }

        /**
         * Constructs a testing video device.
         */
        public VideoDevice.test() {
            base.test();
            name = "Test Video Source";
            device_path = Devices.DEV_NULL;
            var frs = new Gee.TreeSet<Framerate>();
            frs.add(new Framerate(30, 1));
            setups.add(new VideoSetup(VideoSetup.RAW, "I420", 640, 480, frs));
        }

        /**
         * Initializes the video device setups.
         */
        protected override void init_setups() {
            setups = new Gee.TreeSet<VideoSetup>();
        }

        /**
         * Parse an structure in the device.
         * @param structure the structure to parse.
         */
        protected override void parse_structure(Gst.Structure structure) {
            string n = "";
            string f = "";
            int w = -1;
            int h = -1;
            int num =  -1;
            int den = -1;
            n = structure.get_name();
            f = structure.get_string("format");
            if ("image/" in n)
                return;
            if (structure.has_field_typed("width", typeof(int)))
                structure.get_int("width", out w);
            if (structure.has_field_typed("height", typeof(int)))
                structure.get_int("height", out h);
            GLib.Value? v = structure.get_value("framerate");
            if (w < 1 || h < 1 || w < h || v == null)
                return;
            var frs = new Gee.TreeSet<Framerate>();
            if (v.type() == typeof(Gst.ValueList)) {
                for (int i = 0; i < Gst.ValueList.get_size(v); i ++) {
                    var vf = Gst.ValueList.get_value(v, i);
                    if (vf.type() != typeof(Gst.Fraction))
                        continue;
                    parse_fraction(vf, out num, out den);
                    if (num > 0 && den > 0)
                        frs.add(new Framerate(num, den));
                }
            } else if (v.type() == typeof(Gst.Fraction)) {
                parse_fraction(v, out num, out den);
                if (num > 0 && den > 0)
                    frs.add(new Framerate(num, den));
            }
            if (!frs.is_empty)
                setups.add(new VideoSetup(n, f, w, h, frs));
        }

        /* Parses a fraction. */
        private void parse_fraction(GLib.Value frac,
                                    out int num, out int den) {
            num = Gst.Value.get_fraction_numerator(frac);
            den = Gst.Value.get_fraction_denominator(frac);
        }

        /**
         * Compares the two video devices.
         * @param dev the video device to compare with.
         * @return -1 if this device it's better than the one received; 0 if
         *         they are equal; 1 if it's worse.
         */
        public int compare_to(VideoDevice dev) {
            if (!is_test && dev.is_test)
                return -1;
            if (is_test && !dev.is_test)
                return 1;
            var ts = this.setups.first();
            var ds = dev.setups.first();
            return ts.compare_to(ds);
        }

        public new Gst.Element create_element(string name) {
            if (!is_test)
                return base.create_element(name);
            var src = Gst.ElementFactory.make("videotestsrc", name);
            src.set("is-live", true);
            return src;
        }
    }
}
